import json
import pathlib

from jinja2 import Environment, PackageLoader, select_autoescape

import cldfofflinebrowser

__all__ = ['render_to_string', 'render']

env = Environment(
    loader=PackageLoader(cldfofflinebrowser.__name__, 'templates'),
    autoescape=select_autoescape([])
)
env.filters.update(jsondumps=json.dumps, len=lambda v: len(v))


def render_to_string(template, **vars):
    return env.get_template(template).render(**vars)


def render(out, template, **vars):
    out = pathlib.Path(out)
    if out.is_dir():
        out = out / template
    out.write_text(render_to_string(template, **vars), encoding='utf8')



"""
Add a sound file to an offline version
"""


def register(parser):
    pass
    #
    # FIXME: language and parameter ID, sound file path, optionally transcription
    # -> ask Mary what the most likely scenario for adding data would be.
    #


def run(args):
    pass
    #
    # FIXME: must add to HTML as well!
    #



"""
Use the downloadosmtiles command to manage a map tiles.

See http://manpages.ubuntu.com/manpages/xenial/man1/downloadosmtiles.1p.html
"""
import decimal
import pathlib
import itertools
import subprocess
import collections

__all__ = ['CMD', 'TileList']

# The following command will be available in Ubuntu, if libgeo-osm-tiles-perl is installed.
CMD = "downloadosmtiles"


def downloadosmtiles(*args, **opts):
    return subprocess.check_call(
        [CMD] + list(args) + ['--{}={}'.format(k, str(v)) for k, v in opts.items()])


class TileList:
    def __init__(self, p):
        subprocess.call([CMD, '-xyz'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        self.path = pathlib.Path(p)

    def create(self, coords, maxzoom, padding=10):
        """
        Write a list of tiles required to fit all coords in the zoom levels from minzoom to maxzoom.

        :param coords: `list` of (lat, lon) pairs.
        :param maxzoom: `int` maximal required zoom level
        :return:
        """
        if maxzoom > 12:
            # https://operations.osmfoundation.org/policies/tiles/
            raise ValueError('Maxzoom exceeds level allowed for bulk downloading!')

        def _required_tiles():
            return {z: list(t) for z, t in itertools.groupby(self, lambda xyz: int(xyz[2]))}

        lats, lons = list(zip(*coords))
        minlat, maxlat = min(lats), max(lats)
        minlon, maxlon = min(lons), max(lons)
        tiles = collections.defaultdict(set)
        # then enlarge bounding box per zoom level
        for i, zoom in enumerate(range(5, maxzoom + 1)):
            pad = decimal.Decimal(padding / pow(2, i))
            minlonpad, maxlonpad = minlon - pad, maxlon + pad
            minmaxlonpad = None
            if maxlonpad > 180:
                minmaxlonpad = (-180, -360 + maxlonpad)
                maxlonpad = 180
            if minlonpad < -180:
                minmaxlonpad = (360 + minlonpad, 180)
                minlonpad = -180
            downloadosmtiles(
                dumptilelist=self.path,
                lat='{}:{}'.format(minlat - pad, maxlat + pad),
                lon='{}:{}'.format(minlonpad, maxlonpad),
                zoom='{}:{}'.format(zoom, zoom),
            )
            tiles[zoom].update(_required_tiles()[zoom])
            # map crossing date line? add other side
            if minmaxlonpad:
                downloadosmtiles(
                    dumptilelist=self.path,
                    lat='{}:{}'.format(minlat - pad, maxlat + pad),
                    lon='{}:{}'.format(minmaxlonpad[0], minmaxlonpad[1]),
                    zoom='{}:{}'.format(zoom, zoom),
                )
                tiles[zoom].update(_required_tiles()[zoom])
        self.write(tiles)

    def prune(self, d=None):
        """
        Prune the tile list, removing tiles which are already present in the directory `d`.

        :param d: Path to tiles directory.
        :return: `int` number of tiles that are missing in `d`.
        """
        d = pathlib.Path(d) if d else self.path.parent
        missing_tiles = collections.defaultdict(list)
        for zoom, tiles in itertools.groupby(self, lambda xyz: xyz[2]):
            for x, y, z in tiles:
                if not d.joinpath(z, x, '{}.png'.format(y)).exists():
                    missing_tiles[z].append((x, y, z))

        if missing_tiles:
            self.write(missing_tiles)
        return sum(len(v) for v in missing_tiles.values())

    def download(self, d=None):
        d = pathlib.Path(d) if d else self.path.parent
        downloadosmtiles(destdir=d, loadtilelist=self.path)

    def write(self, tiles):
        # Note: We handcraft the rather simple yaml format of tilelists, rather than requiring
        # pyyaml.
        yaml = ['---']
        for zoom, tiles in sorted(tiles.items(), key=lambda i: int(i[0])):
            yaml.append('{}:'.format(zoom))
            for tile in tiles:
                yaml.append('  - xyz:')
                yaml.extend(['      - {}'.format(i) for i in tile])
        self.path.write_text('\n'.join(yaml), encoding='utf8')

    def __iter__(self):
        xyz, index = [None, None, None], -1
        for line in self.path.read_text(encoding='utf8').split('\n'):
            line = line.strip()
            if line == '---':
                continue
            if line.startswith('-'):
                if line.endswith('xyz:'):
                    if index > 0:
                        yield tuple(xyz)
                        index = -1
                else:
                    index += 1
                    xyz[index] = line[1:].strip()
        if index > 0:
            yield tuple(xyz)



import shutil
import pathlib
import collections
from urllib.request import urlretrieve

import rfc3986
from clldutils.path import md5

__all__ = ['PREFERRED_AUDIO', 'download', 'read_media_files', 'get_best_audio']

PREFERRED_AUDIO = collections.OrderedDict([
    ('audio/mpeg', '.mp3'),
    ('audio/wav', '.wav'),
    ('audio/x-wav', '.wav'),
    ('audio/ogg', '.ogg'),
])


def download(cldf, media_row, outdir, fname, media_table='media.csv', md5sum=None):
    target = pathlib.Path(outdir) / fname
    if not target.exists() or (md5sum and md5sum != md5(target)):
        url = cldf.get_row_url(media_table, media_row)
        if isinstance(url, rfc3986.URIReference):
            url = url.unsplit()
        try:  # pragma: no cover
            print(url)
            print(target)
            urlretrieve(url, target)
        except ValueError:
            if cldf.directory.joinpath(url).exists():
                shutil.copy(str(cldf.directory / url), str(target))
            else:  # pragma: no cover
                raise
    return target


def read_media_files(cldf, filter=None):
    """

    :param cldf:
    :param filter:
    :return: A pair of `dict`s, the first mapping media ID to media row, the second mapping form ID\
    to `list` of related media IDs.
    """
    try:
        form_ref = cldf['media.csv', 'formReference']
    except KeyError:
        form_ref = None
    form2media = collections.defaultdict(list)
    media = {}
    try:
        for r in cldf['media.csv']:
            if (filter is None) or filter(r):
                media[r['ID']] = r
                if form_ref:
                    ref = r[form_ref.name]
                    if form_ref.separator and isinstance(ref, list):
                        for rr in ref:
                            form2media[rr].append(r)
                    elif ref:
                        form2media[ref].append(r)
    except KeyError:
        pass
    return media, form2media


def get_best_audio(audios):
    """
    For offline usage, we optimize filesize over widest browser support, so only choose one audio
    file per form.
    :param audios:
    :return:
    """
    if audios:
        pref = {mtype: i for i, mtype in enumerate(PREFERRED_AUDIO)}
        return sorted(
            audios, key=lambda r: pref.get(r['mimetype'], len(pref)))[0]



"""
Create an offline browseable version of a CLDF Wordlist.
"""
import shutil
import pathlib
import itertools
import mimetypes
import collections

from tqdm import tqdm
from cldfbench.cli_util import add_dataset_spec, get_dataset

import cldfofflinebrowser
from cldfofflinebrowser import osmtiles
from cldfofflinebrowser.template import render
from cldfofflinebrowser import media


def register(parser):
    parser.add_argument(
        '--outdir',
        help="Directory in which to create the offline browseable files.",
        default='offline')
    parser.add_argument(
        '--with-tiles',
        help="Also download map tiles (requires {})".format(osmtiles.CMD),
        action='store_true',
        default=False)
    parser.add_argument(
        '--with-audio',
        help="Also download audio files",
        action='store_true',
        default=False)
    parser.add_argument(
        '--include',
        help="Whitespace separated list of parameter IDs",
        type=lambda s: s.split(),
        default=None)
    add_dataset_spec(parser)
    parser.add_argument(
        '--padding',
        default=8,
        help="Padding in degree longitude at zoom level 5 to add to minimal bounding box when "
             "retrieving map tiles.",
        type=int)
    parser.add_argument(
        '--max-zoom',
        default=10,
        help="Maximal zoom level for which to add map tiles (must be < 13)",
        type=int)
    #
    # FIXME: configuration? Name of the media FK column?
    # sorting of markers?
    #


def _recursive_overwrite(src, dest):
    """Copy a folder structure overwriting existing files"""
    if src.is_dir():
        if not dest.exists():
            dest.mkdir(parents=True)
        for f in src.iterdir():
            _recursive_overwrite(f, dest / f.name)
    else:
        shutil.copyfile(str(src), str(dest))


def run(args):
    ds = get_dataset(args)
    cldf = ds.cldf_reader()
    # We expect a list of audio files in a table "media.csv", with a column "mimetype".
    audio, form2audio = media.read_media_files(
        cldf, filter=lambda r: r['mimetype'].startswith('audio/'))
    title_ = cldf.properties['dc:title'].replace('"', '”')
    title = '<div class="truncate">{}.</div>'.format(title_)
    title_tooltip = title_

    outdir = pathlib.Path(args.outdir)
    if not outdir.exists():
        outdir.mkdir()

    for sub in ['tiles', 'static']:
        sub = outdir / sub
        if not sub.exists():
            sub.mkdir()

    for p in pathlib.Path(cldfofflinebrowser.__file__).parent.joinpath('static').iterdir():
        shutil.copy(str(p), str(outdir / 'static' / p.name))

    parameters = {}
    for p in cldf.iter_rows('ParameterTable', 'id', 'name'):
        if (args.include is None) or (p['id'] in args.include):
            p.update(representation=set(), has_audio=False)
            parameters[p['id']] = p

    languages, coords = {}, []
    for p in cldf.iter_rows('LanguageTable', 'latitude', 'longitude', 'id', 'name'):
        for c in ['Latitude', 'Longitude']:
            if c in p:
                del p[c]
        if p['latitude'] is None or p['longitude'] is None:
            continue
        coords.append((p['latitude'], p['longitude']))
        p['latitude'] = float(p['latitude'])
        p['longitude'] = float(p['longitude'])
        languages[p['ID']] = p

    tiles_outdir = outdir / 'tiles'
    _recursive_overwrite(pathlib.Path(__file__).parent.parent / 'tiles', tiles_outdir)
    if args.with_tiles:
        try:
            tiles = osmtiles.TileList(tiles_outdir / 'tilelist.yaml')
        except FileNotFoundError:
            args.log.error(
                'The command {} is not installed on your system. '
                'Either install it or do not use the --with-tiles flag.'.format(
                    osmtiles.CMD))
            return
        tiles.create(coords, args.max_zoom, padding=args.padding)
        missing = tiles.prune()
        if missing:
            args.log.info('Must download {} tiles'.format(missing))
            tiles.download()

    #
    # FIXME: looping over FormTable means we only support Wordlist!
    #
    for pid, forms in tqdm(itertools.groupby(
        sorted(
            cldf.iter_rows('FormTable', 'id', 'languageReference', 'parameterReference', 'form'),
            key=lambda r: (r['parameterReference'], r['id'])),
        lambda r: r['Parameter_ID'],
    )):
        if args.include and (pid not in args.include):
            continue
        data = {
            'languages': {},
            'forms': collections.defaultdict(dict),
        }
        pout = outdir / 'parameter-{}'.format(pid)
        if not pout.exists():
            pout.mkdir()

        for form in forms:
            if form['languageReference'] not in languages:
                continue
            data['forms'][form['languageReference']] = {
                'form': form['form'],
                'audio': None,
            }
            data['languages'][form['languageReference']] = languages[form['languageReference']]
            parameters[pid]['representation'].add(form['languageReference'])
            if 'Audio_Files' in form:
                # Audio files may either be linked via a list-valued foreign key column
                # "Audio_Files" ...
                audio_files = [audio[aid] for aid in form['Audio_Files'] if aid in audio]
            else:
                # ... or via a formReference in the media table:
                audio_files = form2audio.get(form['id'], [])
            audio_file = media.get_best_audio(audio_files)
            if audio_file and args.with_audio:
                suffix = media.PREFERRED_AUDIO.get(audio_file['mimetype']) \
                    or mimetypes.guess_extension(audio_file['mimetype']) \
                    or '.bin'
                data['forms'][form['languageReference']]['audio'] = {
                    'name': media.download(
                        cldf,
                        audio_file,
                        pout,
                        '{}{}'.format(form['languageReference'], suffix)).name,
                    'mimetype': audio_file['mimetype'],
                }
                parameters[pid]['has_audio'] = True

        render(
            pout,
            'data.js',
            data=data,
            options={'minZoom': 0, 'maxZoom': args.max_zoom})
        render(
            pout / 'index.html',
            'parameter.html',
            parameter=parameters[pid],
            index=False,
            data=data,
            parameters=parameters.items(),
            title_tooltip=title_tooltip,
            title=title,
        )

    for c in ['forms', 'languages']:
        if c in data:
            del data[c]
    data['index'] = True
    data['languages'] = {}
    for k, v in languages.items():
        data['languages'][k] = {
            'Name': v['Name'],
            'latitude': v['latitude'],
            'longitude': v['longitude'],
        }

    render(
        outdir,
        'data.js',
        data=data,
        options={'minZoom': 0, 'maxZoom': args.max_zoom})
    render(
        outdir,
        'index.html',
        parameters=parameters.items(),
        index=True,
        data=data,
        has_any_audio=any(p['has_audio'] for p in parameters.values()),
        title_tooltip=title_tooltip,
        title=title,
    )
